/***************************************************************
 ************************* A TPP Project ***********************
 ******************** Visit http://tpp.dhs.org *****************
 ***************************************************************
translate.cpp

Author(s): Paul Fleischer (polle@denmark.crosswinds.net)
Created: 20. June 1999

Copyright(c) 1999 The AKOS team.
All Rights Reserved.
****************************************************************/

#include "main.h"


/**********************************************************************
-----------------------------------------------------------------------
bool
translate(char *string, CODE_BLCOK *code)
-----------------------------------------------------------------------
Info: Convert the 'string'. Performs all calculations in the string,
      ie. resolves the symbols "+","-","*","/".
      Parantheses supported!
      In the future, this function will also resolve variables
      
      translate() uses transmul() for doing the actual calculations.

In: 'string'; the code string to convert (or translate).
    'code'; Currently only used to get the current line numer

Out: 'string'; is changed directly.

Return: bool; true if everything goes well, false otherwise.

BUGS: All variables are static allocated with a length of 255. This means
      that the string never may be larger than 255.
***********************************************************************/
bool
translate(char *string, CODE_BLOCK *code)
{
  unsigned int i=0;          //You guessed it, a counter variable
  bool end=false;

  //First run through the string to find any paranteses '(', ')' and resolve them.
  while(!end)
    {
      unsigned int stop;       //Used to find the ending paranthese ')'.
      char str[255]; //Used for pasing the string between the parantheses.

      //Run through the string, until we encounter a '('
      for(i=0; (i<strlen(string)) && 
	    (string[i] != '(')
	    ; i++);
      
      if(i >= strlen(string)) end = true;
    
      if(!end)  //If we're not at the end of the string, we can go on.
	{
	  int pcount=0;
	 
	  //Search for the ending ')' (and NOT just a ')')
	  for(stop=i; (stop<strlen(string)) &&
		!((string[stop] == ')') && (pcount<=1))
		; stop++)
	    {
	      if(string[stop] == '(') pcount++;
	      if(string[stop] == ')') pcount--;
	    }

	  if(stop >= strlen(string)) end = true;

	  //If we're at the end of the string, we have not found ')'
	  //and need to return with an error.
	  if(end) 
	    {
	      return false;
	    }

	  //Find the string between the parantheses.
	  strncpy(str, string+i+1, stop-i-1);
	  str[stop-i-1] = 0;

	  //Send it to translate() (pretty clever, eh?)
	  if(!translate(str, code)) return false;
	  
	  //Now we have to put 'str' into 'string'
	  //Well, we first copy all the stuff into 'tmpstr', 
	  //and then to 'string'
	  char tmpstr[255];

	  strncpy(tmpstr, string, i);
	  tmpstr[i] = 0;  //Put in the missing '\0' 
	                  //(since strncpy doesn't do it)

	  strncat(tmpstr, str, strlen(str));
	  tmpstr[strlen(str)+i] = 0;  //Put in the missing '\0'

	  strncat(tmpstr, string+stop+1, strlen(string)-stop-1);
	  tmpstr[(strlen(string)-stop)+i+strlen(str)-1] = 0; //Put in the missing \0

	  strncpy(string, tmpstr, strlen(tmpstr));
	  string[strlen(tmpstr)] = 0;
	}
    }

  /*
  //Now we search for any possible variables, and resolve them
  end = false;
  while(!end)
    {
      for(i=0; (i<strlen(string)) &&
	    (!isalpha(string[i]))
	    ; i++);


      //Now we have to find out the end of the possible variable string
      int l=i;
      
      for(;(l<strlen(string)) &&
	    (isalpha(string[l]))
	    ; l++);

      
      if(!isalpha(string[i])) end = true;

      if(!end) 
	{
	  
	  char *tmpstr;
      
	  strncpy(tmpstr, string+i, l-i);
	  tmpstr[l-i] = 0;

	  //Check if the variable exists..
	  if(!varExists(code, tmpstr))
	    {
	      //If not emit an error(or would it be warning?) message.
	      char errormsg[255];
	      sprintf(errormsg, "Variable '%s' not defined!", tmpstr);
	      emitError(code->name, code->linePointer, errormsg);
	      strcpy(tmpstr, "0");
	    }
	  else
	    {
	      char tmpvar[255];
	      //If the variable exists, look up it's value and put it into tmpstr
	      getValue(code, tmpvar, tmpstr);
	      strcpy(tmpstr, tmpvar);
	    }
	  
	  //Now we just need to substitute the variable name with the value in tmpstr

	  char newString[255];
	  
	  strncpy(newString, string, i);
	  newString[i] = 0;
	  strcat(newString, tmpstr);

	  if(l<strlen(string)-1)
	    strcat(newString, string+l);
 
	  strcpy(string, newString);
	}
    }
  */

  //Run through the string and find all * and /, and calculate their result
  end = false;
  while(!end)
    {
      //Run through the string, until we encounter a calculation symbol
      for(i=0; (i<strlen(string)) && 
	    (string[i] != '*') &&
	    (string[i] != '/')
	    ; i++);
      
      if(i >= strlen(string)) end = true;

      //Determine which calculation has to be performed, and call transmul,
      //which does the actual calculation

      if(string[i] == '*') 
	if(!transmul(string, i, '*', code)) return false;
      
      if(string[i] == '/')
	if(!transmul(string, i, '/', code)) return false;
    }

  //Now run through the string once more, to calculate all + and -.
  end = false;
  i = 0; //We cannot reset i in every for-loop, since there might be a -
         //on the line, to represent a negative value.
  while(!end)
    {
      for(; (i<strlen(string)) && 
	    (string[i] != '+') &&
	    !((string[i] == '-') && (string[i-1] != ' '))
	    ; i++);
      
      if(i >= strlen(string)) end = true;

      if(string[i] == '+')
	{
	  if(!transmul(string, i, '+', code)) return false;
	  i = 0;
	}

      if(string[i] == '-')
	{
	  if(i != 0)
	    {
	      if(!transmul(string, i, '-',code)) return false;
	    }
	  else
	    i++;
	}
    }

  //Now comes the last part, resolve expressions like: "1==3"
  end = false;
  while(!end)
    {
      for(i=0; (i<strlen(string)) && 
	    !((string[i] == '=') && (string[i+1] == '=')) &&
	    !((string[i] == '!') && (string[i+1] == '=')) &&
	    !((string[i] == '&') && (string[i+1] == '&')) &
	    !((string[i] == '|') && (string[i+1] == '|')) &&
	    !(string[i] == '>') &&
	    !(string[i] == '<')
	    ; i++);
      
      if(i >= strlen(string)) end = true;

      if( (string[i] == '=') && (string[i+1] == '=') )
	if(!transmul(string, i, 'e', code)) return false;

      if( (string[i] == '!') && (string[i+1] == '=') )
	if(!transmul(string, i, 'n', code)) return false;
    
      if( (string[i] == '&') && (string[i+1] == '&') )
	if(!transmul(string, i, 'a', code)) return false;

      if( (string[i] == '|') && (string[i+1] == '|') )
	if(!transmul(string, i, 'o', code)) return false;
  
      if( string[i] == '>' )
	if(!transmul(string, i, 'g', code)) return false;
      
      if( string[i] == '<' )
	if(!transmul(string, i, 'l', code)) return false;
    }

  return true;
}

/**********************************************************************
-----------------------------------------------------------------------
bool
transmul(char *string, int i, char mode, CODE_BLOCK *code)
-----------------------------------------------------------------------
Info: Used by translate for doing calculations, should not be called
      from anywhere else.

In: 'string'; the string which contains the "commandline". 
    'i'; where in the string the calculation symbol is.
    'mode'; what type of calculation to perform(case sensitive):
            * multiply
	    / divide
	    + plus
	    - minus
	    e Equal(==)
	    n Not equal(!=)
	    g Greater then(>)
	    l Less then(<)
	    a And(&&)
	    o Or(||)
     'code'; only used for getting the current line for error messages

Out: 'string' is changed...

Return: bool; false if an error occured, true otherwise.
***********************************************************************/
bool
transmul(char *string, unsigned int i, char mode, CODE_BLOCK *code)
{
  //Everything is static right now, but it would be nicer if it were Dynamic
  char firstArg[255];
  char secondArg[255];

  int firstInt;
  int secondInt;
  char output[255];
  bool argsOk=true;
  int start, stop;

  if(i<strlen(string)) //We just wanna be sure, that we have not reached
                       //The end of the string (which only occurs on invalid
                       //calls, but we still have to be sure ;)
    {
      //Find the two arguments.
      //The first one is the one before the symbol, and the second after.

      unsigned int l=i-1;

      //Now find the start of the first argument.
      //This is done by going backwards, until the character is no longer
      //a digit
      for(; ((isdigit(string[l]) !=0) || (string[l] == '-'))
	    && (l>0)
	    ; l--);
      
      
      //Now we just have to copy the argument from the string
      start = l;   //We have to store the position of the first argument
      //to be able to put the string back together again
      strncpy(firstArg, string+l, i-(l));
      firstArg[(i-l)] = 0; //strncpy doesn't put the \0 in the end 
                               //of the string, so we have to do it 

      //If the first argument has zero length, emit an error message
      if(firstArg[0] == 0)
	{
	  argsOk = false;
	}

      //Up next: The second argument ;)

      //If the mode is e(==) or n(!=), there are two chars and  we have to 
      //skip both
      if( (mode == 'e') || (mode == 'n') || (mode == 'a') || (mode == 'o') )
	i++;
      
      l=i+1;
      //In case of a - sign before the actual number, interpret it as
      //a negative value
      //This is done, by increasing l, so that the - sign isn't handled
      //by the for() instruction below.
      if( (string[l] == '-') ) l++;

      //Find the second argument, by going forward through the string,
      //until we reach a character not being a digit
      for(; (isdigit(string[l]) != 0)
	    && (l<strlen(string))
	    ; l++);
      
  
      stop = l; //Store the second arguments end (we'll need it when putting
                //the string back together again)
      strncpy(secondArg, string+i+1, l-i);
      secondArg[(l-i)-1] = 0;  //strncpy doesn't put the \0 in the end of
                               //the string, so we'll have to do it.
      

      //If the second argument has zero length, emit an error message
      if(secondArg[0] == 0)
	{
	  argsOk = false;
	}
    }

  //If both arguments are given, we can now go on to the calculation, and
  //put the string back together again, and ship it out of the function ;)
  if(argsOk)
    {
      //Convert the argument to integers
      firstInt = strtol(firstArg, NULL, 10);
      secondInt = strtol(secondArg, NULL, 10);      

      //Now do the actual calculation (the simplest part)
      //The calculation that has to be performed, depends on the mode char
      if(mode == '*')
	{
	  firstInt = firstInt * secondInt;
	}

      if(mode == '/')
	{ 
	  if(secondInt == 0) //We don't want any run-time errors, so we do the
	                     //divide by zero checking ourself.
	    {
	      //printf("ERROR: '/', Divide by zero!\n");
	      emitError("Unknown", 0, "Divide by Zero!");
	      return false;
	    }
	    
	  firstInt = firstInt/secondInt;
	}

      if(mode == '+')
	firstInt = firstInt + secondInt;

      if(mode == '-')
	firstInt = firstInt - secondInt;

      if(mode == 'e')
	{
	  if(firstInt == secondInt)
	    firstInt = 1;
	  else
	    firstInt = 0;
	} 

      if(mode == 'n')
	{
	  if(firstInt != secondInt)
	    firstInt = 1;
	  else
	    firstInt = 0;
	}

      if(mode == 'g')
	{
	  if(firstInt > secondInt)
	    firstInt = 1;
	  else
	    firstInt = 0;
	}

      if(mode == 'l')
	{
	  if(firstInt < secondInt)
	    firstInt = 1;
	  else
	    firstInt = 0;
	}

      if(mode == 'a')
	{
	  if((firstInt == 1) && (secondInt == 1))
	    firstInt = 1;
	  else
	    firstInt = 0;
	}

      if(mode == 'o')
	{
	  if((firstInt == 1) || (secondInt == 1))
	    firstInt = 1;
	  else
	    firstInt = 0;
	}

      //Take the first part of the string (until the start of the first arg.)
      strncpy(output, string, start);
      output[start] = 0; //Put in the missing \0

      //Put in the calculated number
      sprintf(output, "%s%i", output, firstInt);

      //And unless we are at the end of the string, put the rest at the end
      //(From the end of the second argument)
      if((unsigned int)stop < strlen(string))
	  strcat(output, string+stop);
  
      //Now copy it back to the input string...
      strcpy(string, output);

      //...and we are ready to return the OK value (which is 'true' right now)
      return true;
    } else return false;

  return false;
}








